#ifndef APPLICATION_HPP
#define APPLICATION_HPP

#include <libnotify/notify.h>

#include <QObject>
#include <QApplication>
#include <QHostAddress>
#include <QSystemTrayIcon>
#include <QIcon>
#include <QPixmap>
#include <QMenu>
#include <QAction>
#include <QList>
#include <QMap>
#include <QTcpSocket>
#include <QTextCodec>
#include <QStandardPaths>
#include <QDir>
#include <QSettings>
#include <QProcess>
#include <QLocalSocket>

#include <algorithm>
#include <iostream>
#include <stdio.h>

#include <lektor/common.h>
LKT_PUSH
LKT_IGNORE_WARN_WRITE_STRING
#include <lektor/icon.xpm>
LKT_POP

static inline const int UTF8_MIB = 106;

/* Hop();; We wrap the glib types */
#define URGENCY(x) x = NOTIFY_URGENCY_ ##x,
enum Urgency {
    URGENCY(LOW)
    URGENCY(NORMAL)
    URGENCY(CRITICAL)
};
#undef URGENCY

static inline const char *
__urgencyToString(Urgency urg)
{
    switch (urg) {
    case Urgency::LOW:
        return "LOW";
    case Urgency::NORMAL:
        return "NORMAL";
    case Urgency::CRITICAL:
        return "CRITICAL";
    default:
        return "UNKNOWN";
    }
}

#if defined (Q_OS_LINUX)
/* Restrict libnotify (thus glib...) only to that function */
static void
sendNotification(const char *name, const char *msg, enum Urgency urgency)
{
    GdkPixbuf *gtk_icon       = gdk_pixbuf_new_from_xpm_data(const_cast<const char **>(icon));
    NotifyNotification *notif = notify_notification_new (name, msg, NULL);
    notify_notification_set_icon_from_pixbuf(notif, gtk_icon);
    notify_notification_set_urgency(notif, static_cast<NotifyUrgency>(urgency));
    notify_notification_show(notif, NULL);
    g_object_unref(G_OBJECT(notif));
    g_object_unref(gtk_icon);
    fprintf(stdout, "# %s (%s) -> %s\n", name, __urgencyToString(urgency),
            QString(msg).replace('\n', "\\n").toStdString().c_str());
    fflush(stdout);
}
#else
#error "System is not supported yet, you need to write a 'sendNotification(const char *, const char *, enum Urgency)' function for that system"
#endif

/* Config class, the ~/.config/lektor/klkt.ini file */
class Config final
{
private:
    static inline const QString opt_host = "lektord/host";
    static inline const QString opt_port = "lektord/port";

    QString m_host;
    qint16  m_port;

public:
    Config(QApplication *app, const QString &file) {
        QSettings settings(file, QSettings::IniFormat, app);
        m_host = settings.value(opt_host, "127.0.0.1").toString();
        m_port = settings.value(opt_port, "6600").toInt();

        settings.setValue(opt_host, m_host);
        settings.setValue(opt_port, m_port);
    }

    qint16
    getPort(void) const          {
        return m_port;
    }
    bool
    isUnix(void) const           {
        return m_port <= 0;
    }
    const  QString &
    getHost(void) const {
        return m_host;
    }
};

/* Socket, can be unix socket or tcp socket. Because QIODevice don't have all that what we need */
class Socket final : public QObject
{
    Q_OBJECT
private:
    enum Type {
        Unix,
        Tcp,
    };

    /* Can be tcp or unix socket */
    QTcpSocket   m_tcp;
    QLocalSocket m_unix;
    Type         m_type;

    /* Where to connect */
    QString      m_host;
    qint16       m_port;

public slots:
    /* Re-emit signals */
    inline void
    triggetConnected(void)    {
        emit connected();
    }
    inline void
    triggetDisconnected(void) {
        emit disconnected();
    }
    inline void
    triggetReadyRead(void)    {
        emit readyRead();
    }

signals:
    /* Used signals from TCP and Unix sockets */
    void connected(void);
    void disconnected(void);
    void readyRead(void);

public:
    Socket(const QString &host, qint16 port) {
        reset(host, port);
    }

    void
    reset(const QString host, qint16 port) {
        m_host = host;
        m_port = port;

        /* Unix specific stuff */
        if (port <= 0) {
            m_type = Type::Unix;
            connect(&m_unix, &QLocalSocket::connected,    this, &Socket::triggetConnected);
            connect(&m_unix, &QLocalSocket::disconnected, this, &Socket::triggetDisconnected);
            connect(&m_unix, &QLocalSocket::readyRead,    this, &Socket::triggetReadyRead);
        }

        /* TCP specific stuff */
        else {
            m_type = Type::Tcp;
            connect(&m_tcp, &QTcpSocket::connected,    this, &Socket::triggetConnected);
            connect(&m_tcp, &QTcpSocket::disconnected, this, &Socket::triggetDisconnected);
            connect(&m_tcp, &QTcpSocket::readyRead,    this, &Socket::triggetReadyRead);
        }
    }

#define __DIFFERENTIATE(unix, tcp) {            \
        switch (m_type) {                       \
        case Type::Unix:  unix; break;          \
        case Type::Tcp:   tcp;  break;          \
        }                                       \
        sendNotification("Klkt fatal error",    \
                         "Unknown socket type", \
                         Urgency::CRITICAL);    \
        ::exit(EXIT_FAILURE);                   \
    }

    inline bool
    isValid(void)          __DIFFERENTIATE({ return m_unix.isValid(); }, { return m_tcp.isValid(); })
    inline QByteArray
    readAll(void)    __DIFFERENTIATE({ return m_unix.readAll(); }, { return m_tcp.readAll(); })
    inline qint64
    write(const char *d) __DIFFERENTIATE({ return m_unix.write(d);  }, { return m_tcp.write(d);  })
    inline void
    close(void)            __DIFFERENTIATE({ return m_unix.close();   }, { return m_tcp.close();   })

    inline void
    doConnect(void) __DIFFERENTIATE({
        return m_unix.connectToServer(m_host);
    }, {
        return m_tcp.connectToHost(m_host, m_port);
    })

    inline bool
    waitForConnected(int ms = 30000) __DIFFERENTIATE({
        return m_unix.waitForConnected(ms);
    }, {
        return m_tcp.waitForConnected(ms);
    })

    inline bool
    waitForDisconnected(int ms = 30000) __DIFFERENTIATE({
        return m_unix.waitForDisconnected(ms);
    }, {
        return m_tcp.waitForDisconnected(ms);
    })
#undef __DIFFERENTIATE
};

/* Used to generate a notification for the state of lektord and get the currently playing kara */
class Status final : public QWidget
{
    Q_OBJECT
private:
    Socket                  m_sock;
    QString                 m_pending_data;
    QString                 m_state;
    QMap<QString, QString>  m_current_kara;
    QMetaObject::Connection m_connect_read;
    bool                    m_deletable = false;
    bool                    m_send_notif = true;

    static inline const char *StatusString  = "status\n";
    static inline const char *CurrentString = "currentsong\n";

    void
    sendNotification(const char *title, const char *msg, Urgency urg) {
        if (!m_send_notif)
            return;
        ::sendNotification(title, msg, urg);
    }

private slots:
    void
    connected(void) {
        m_sock.write(StatusString);
    }

    void
    disconnected(void) {
        if (m_deletable)
            return;

        if (m_state == "play") {
            // m_current_kara["category"] + " - " + m_current_kara["language"] + "\n" +
            QString kara = m_current_kara["source"] + "\n"
                           + m_current_kara["type"] + " - " + m_current_kara["title"]
                           + " [" + m_current_kara["author"] + "]";
            sendNotification("lektord is playing", kara.toStdString().c_str(), Urgency::LOW);
            emit changeCurrentKara(kara);
            emit changeStatus(Play);
        }

        else if (m_state == "stop") {
            sendNotification("lektord is stopped", "lektord is not playing", Urgency::LOW);
            emit changeCurrentKara("lektord is stopped");
            emit changeStatus(Stop);
        }

        else if (m_state == "pause") {
            sendNotification("lektord is paused", "lektord is not playing", Urgency::LOW);
            emit changeCurrentKara("lektord is paused");
            emit changeStatus(Pause);
        }

        else {
            sendNotification("lektord error",
                             tr("lektord has the following unkown state: '%1'").arg(m_state).toStdString().c_str(),
                             Urgency::CRITICAL);
            emit changeCurrentKara(tr("Unknown state: '%1'").arg(m_state));
            emit changeStatus(Unknown);
        }

        m_deletable = true;
    }

    void
    readCurrent(void) {
        QByteArray got    = m_sock.readAll();
        m_pending_data   += QTextCodec::codecForMib(UTF8_MIB)->toUnicode(got);
        QStringList lines = m_pending_data.split("\n");
        int return_position;

        for (auto &line : lines) {
            return_position = line.length() + 1;
            m_pending_data  = m_pending_data.right(m_pending_data.length() - return_position);
            if (line == "OK" || line.left(3) == "ACK") {
                m_sock.close();
                return;
            }

            /* Decode line */
#define DECODE(what)                                           \
    if (line.startsWith(what ": ", Qt::CaseInsensitive)) {     \
        line.remove(0, static_cast<int>(strlen(what ": ")));   \
        m_current_kara.insert(what, line);                     \
    }
            DECODE("title")
            DECODE("author")
            DECODE("source")
            DECODE("type")
            DECODE("category")
            DECODE("language")
#undef DECODE
        }
    }

    void
    readStatus(void) {
        QByteArray got    = m_sock.readAll();
        m_pending_data   += QTextCodec::codecForMib(UTF8_MIB)->toUnicode(got);
        QStringList lines = m_pending_data.split("\n");
        int return_position;

        for (const auto &line : lines) {
            return_position = line.length() + 1;
            m_pending_data  = m_pending_data.right(m_pending_data.length() - return_position);
            if (line == "OK" || line.left(3) == "ACK") {
                /* Read the current song if needed */
                if (m_state == "play") {
                    QWidget::disconnect(m_connect_read);
                    m_pending_data = "";
                    m_connect_read = QWidget::connect(&m_sock, &Socket::readyRead, this, &Status::readCurrent);
                    m_sock.write(CurrentString);
                } else
                    m_sock.close();
                return;
            }

            QStringList play_state = line.split(": ");
            if (play_state.length() == 2 && play_state.at(0) == "state")
                m_state = play_state.at(1);
        }
    }

public:
    Status(const QString &host, qint16 port, bool send_notif = true)
        : m_sock(host, port), m_send_notif(send_notif) {
        fprintf(stdout, "Query status%s\n", send_notif ? "" : " (no notification)");
        m_sock.doConnect();
        QWidget::connect(&m_sock, &Socket::connected,    this, &Status::connected);
        QWidget::connect(&m_sock, &Socket::disconnected, this, &Status::disconnected);
        m_connect_read = QWidget::connect(&m_sock, &Socket::readyRead, this, &Status::readStatus);
        fflush(stdout);
    }

    Status &operator=(Status &) = delete;

    inline bool
    isDeletable(void) const   {
        return m_deletable;
    }
    inline void
    waitForDisconnected(void) {
        m_sock.waitForDisconnected(-1);
    }

    enum {
        Pause,
        Play,
        Stop,
        Unknown,
    };

signals:
    void changeCurrentKara(QString);
    void changeStatus(int);
};

/* Send commends to lektord */
class Commander final : public QObject
{
    Q_OBJECT
private:
    Socket      m_sock;
    const char *m_cmd;

    static inline const char *CMD_NEXT       = "next\n";
    static inline const char *CMD_PREVIOUS   = "previous\n";
    static inline const char *CMD_SHUFFLE    = "shuffle\n";
    static inline const char *CMD_FORCE_PLAY = "play\n";
    static inline const char *CMD_TOGGLE     = "pause\n";

    void
    doCommand(void) {
        fprintf(stdout, "Send command %s", m_cmd);
        m_sock.doConnect();
        connect(&m_sock, &Socket::connected,    this, [&]() {
            m_sock.write(m_cmd);
        });
        connect(&m_sock, &Socket::disconnected, this, [&]() {
            emit finished();
        });
        connect(&m_sock, &Socket::readyRead,    this, [&]() {
            QStringList lines = QTextCodec::codecForMib(UTF8_MIB)->toUnicode(m_sock.readAll()).split("\n");
            for (auto &line : lines) {
                if (line == "OK" || line.left(3) == "ACK") {
                    m_sock.close();
                    emit finished();
                    return;
                }
            }
        });
        fflush(stdout);
        m_sock.waitForDisconnected();
    }

    Commander(const QString &host, qint16 port, const char *command)
        : m_sock(host, port), m_cmd(command) {
        doCommand();
    };

    Commander(const QString &host, qint16 port)
        : m_sock(host, port), m_cmd(nullptr)
    {}

public slots:
#define DECLARE_CMD(name) static void name(const QString &host, qint16 port) { Commander cmd(host, port, CMD_##name); }
    DECLARE_CMD(NEXT)
    DECLARE_CMD(PREVIOUS)
    DECLARE_CMD(SHUFFLE)
    DECLARE_CMD(TOGGLE)
    DECLARE_CMD(FORCE_PLAY)

    void handleCorrectPlay(int sta) {
        switch (sta) {
        case Status::Play:
        case Status::Pause:
            m_cmd = CMD_TOGGLE;
            doCommand();
            break;

        default:
            m_cmd = CMD_FORCE_PLAY;
            doCommand();
            break;
        }
    }

    static void
    PLAY(const QString &host, qint16 port) {
        std::cout << "Asked a PLAY command" << std::endl;
        Status sta(host, port, false);
        Commander cmd(host, port);
        connect(&sta, &Status::changeStatus, &cmd, &Commander::handleCorrectPlay);
        sta.waitForDisconnected();
    }
#undef DECLARE_CMD

signals:
    void finished(void);
};

/* The application */
class Application final : public QApplication
{
    Q_OBJECT
private:
    bool            m_exit;
    QPixmap         m_xpm;
    QIcon           m_icon;
    QSystemTrayIcon m_tray_icon;
    QMenu           m_menu;

    bool            m_already_red  = false;
    bool            m_is_connected = false;
    QString         m_pending_data;

    QString         m_host = "127.0.0.1";
    qint16          m_port = 6600;
    QString         m_config_file;

    Socket          m_sock;

    QList<Status *>  m_garbage;

    static inline const char *ListenString = "idle player\n";

protected slots:
    void
    notifyExit(void) {
        garbageCollect(true);

        if (m_exit)
            return;

        sendNotification("Klkt quitting", "Klkt will stop", Urgency::LOW);
        m_exit = true;
        if (m_sock.isValid())
            m_sock.close();

        quit();
    }

    void
    disconnected(void) {
        if (m_exit || !m_is_connected)
            return;

        sendNotification("Connexion lost", "Connexion to lektord has been lost!", Urgency::CRITICAL);
        m_pending_data = "";
        m_is_connected = false;
    }

    void
    reconnect(void) {
        Config conf(this, m_config_file);
        if (conf.isUnix()) {
#if defined (Q_OS_MSDOS) || defined (Q_OS_WIN32) || defined (Q_OS_WINCE) || defined (Q_OS_WIN) || defined (Q_OS_WINRT)
            sendNotification("Klkt fatal error", "The 'unix' socket type is not available on your platform",
                             Urgency::CRITICAL);
#endif
        }

        m_host = conf.getHost();
        m_port = conf.getPort();
        m_sock.reset(m_host, m_port);
        m_sock.doConnect();
        if (!m_sock.waitForConnected(50000))
            sendNotification("Klkt not connected", "Connexion to lektord failed", Urgency::CRITICAL);
        else
            m_is_connected = true;
    }

    void
    read(void) {
        QByteArray got = m_sock.readAll();
        m_pending_data = QTextCodec::codecForMib(UTF8_MIB)->toUnicode(got);

        if (! m_already_red) {
            m_already_red = true;
            m_sock.write(ListenString);
        }

        for (const auto &line : m_pending_data.split("\n")) {
            if (line.left(2) == "OK")
                queryStatus();
        }
    }

    void
    connected(void) {}

private:
    void
    queryStatus(void) {
        garbageCollect();
        Status *ptr_status = new Status(m_host, m_port);
        connect(ptr_status, &Status::changeCurrentKara, this, [&](QString kara) {
            m_tray_icon.setToolTip(kara);
        });
        m_garbage.push_back(ptr_status);
    }

    void
    garbageCollect(bool force = false) {
        std::transform(m_garbage.begin(), m_garbage.end(),
        m_garbage.begin(), [&force] (Status * p_sta) -> Status* {
            if (force || (p_sta && p_sta->isDeletable())) {
                delete p_sta;
                return nullptr;
            } else
                return p_sta;
        });

        /* Should be enaugh to delete all used status sockets */
        while (m_garbage.length() > 0 && m_garbage.front() == nullptr)
            m_garbage.pop_front();
    }

    // Qt::GenericConfigLocation
public:
    Application(int argc, char **argv)
        : QApplication(argc, argv), m_exit(false), m_xpm(icon), m_icon(m_xpm), m_tray_icon(m_icon),
          m_config_file(QStandardPaths::writableLocation(QStandardPaths::GenericConfigLocation)),
          m_sock(m_host, m_port) {
        notify_init("Klkt");

        {
            QDir config_dir(m_config_file);
            if (!config_dir.mkpath(m_config_file)) {
                sendNotification("Klkt fatal error",
                                 tr("Failed to create the Klkt config dir %1").arg(m_config_file).toStdString().c_str(),
                                 Urgency::CRITICAL);
                ::exit(EXIT_FAILURE);
            }
            m_config_file += "/lektor/klkt.ini";
            QFile conf_file(m_config_file);
            if (!conf_file.open(QIODevice::ReadWrite | QIODevice::Text)) {
                sendNotification("Klkt fatal error",
                                 tr("Failed to create Klkt config file %1").arg(m_config_file).toStdString().c_str(),
                                 Urgency::CRITICAL);
                ::exit(EXIT_FAILURE);
            }
            conf_file.close();
        }

        {
            m_menu.addSection("Playback");
            QAction *playback_play  = m_menu.addAction(QIcon(":/icons/play.png"),       "Play/Pause");
            QAction *playback_next  = m_menu.addAction(QIcon(":/icons/next.png"),       "Next");
            QAction *playback_prev  = m_menu.addAction(QIcon(":/icons/previous.png"),   "Previous");
            QAction *playback_shuf  = m_menu.addAction(QIcon(":/icons/shuffle.png"),    "Shuffle");
            m_menu.addSection("System");
            QAction *action_reco    = m_menu.addAction("Reconnect");
            QAction *action_restart = m_menu.addAction("Restart");
            QAction *action_quit    = m_menu.addAction("Quit");
            m_tray_icon.setContextMenu(&m_menu);
            m_tray_icon.setToolTip("The lektord mk7 system tray client");

            playback_play->setIconVisibleInMenu(true);
            playback_prev->setIconVisibleInMenu(true);
            playback_next->setIconVisibleInMenu(true);
            playback_shuf->setIconVisibleInMenu(true);

            connect(playback_play,  &QAction::triggered, this, [&]() {
                Commander::PLAY(m_host, m_port);
            });
            connect(playback_next,  &QAction::triggered, this, [&]() {
                Commander::NEXT(m_host, m_port);
            });
            connect(playback_prev,  &QAction::triggered, this, [&]() {
                Commander::PREVIOUS(m_host, m_port);
            });
            connect(playback_shuf,  &QAction::triggered, this, [&]() {
                Commander::SHUFFLE(m_host, m_port);
            });

            connect(action_quit,    &QAction::triggered, this, &Application::notifyExit);
            connect(action_reco,    &QAction::triggered, this, &Application::reconnect);
            connect(action_restart, &QAction::triggered, this, &Application::restart);
            connect(&m_tray_icon,   &QSystemTrayIcon::activated, &m_menu, [&]() {
                m_menu.popup(QCursor::pos());
            });
        }

        m_tray_icon.show();

        connect(&m_sock, &Socket::connected,    this, &Application::connected);
        connect(&m_sock, &Socket::disconnected, this, &Application::disconnected);
        connect(&m_sock, &Socket::readyRead,    this, &Application::read);

        reconnect();
    }

    void
    restart(void) const {
        sendNotification("Klkt restart", "Klkt will restart", Urgency::LOW);
        QStringList args{QApplication::applicationFilePath()};
        QProcess::startDetached(args[0], args);
        ::exit(EXIT_SUCCESS);
    }
};

#endif // APPLICATION_HPP
